@@ -1,38 +1,11 @@
-#
-#  Copyright (c) 1997-2002 The Protein Laboratory, University of Copenhagen
-#  All rights reserved.
-#
-#  Redistribution and use in source and binary forms, with or without
-#  modification, are permitted provided that the following conditions
-#  are met:
-#  1. Redistributions of source code must retain the above copyright
-#     notice, this list of conditions and the following disclaimer.
-#  2. Redistributions in binary form must reproduce the above copyright
-#     notice, this list of conditions and the following disclaimer in the
-#     documentation and/or other materials provided with the distribution.
-#
-#  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
-#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-#  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
-#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
-#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
-#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
-#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
-#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#  SUCH DAMAGE.
 #  Created by:
 #     Anton Berezin  <tobez@tobez.org>
 #     Dmitry Karasik <dk@plab.ku.dk> 
 #     Vadim Belman   <voland@plab.ku.dk>
-#
-#  $Id$
-#
-# max error is APC058
 
 package Prima::Gencls;
 use strict;
+use warnings;
 use Exporter;
 use vars qw( @ISA @EXPORT);
 use vars qw(
@@ -193,7 +166,7 @@ sub init_variables
 	$genDyna   = 0;		# generate exportable files
 	$genObject = 0;		# generate object files
 	$genPackage = 0;		# generate object or package files
-	@includeDirs = qw(./);	# where to search .clses
+	@includeDirs = qw(.);	# where to search .clses
 	%definedClasses = ();	# hash for class search
 	@allVars = ();		# all variables, in order of definition
 	%allVars = ();		# hash for var search
@@ -218,9 +191,9 @@ sub init_variables
 
 
 	%mapTypes = (
-		"int" => "int", "Bool" => "Bool", "Handle" => "Handle", "long" => "int", "short" => "int",
-		"char" => "int", "U8" => "int", "UV" => "UV", 
-		"double" => "double", "Color" => "UV", "SV"=> "SV", "HV"=> "HV");
+		( long => 'int', U8 => 'int', char => 'int' ),
+		map {( $_, $_ )} qw(int Bool Handle UV Color double SV HV),
+	);
 
 	%typedefs = ();
 
@@ -229,6 +202,7 @@ sub init_variables
 		# 0         1            2           3        4           5        6        7       8
 		'int'     => ['int',    'SvIV',      'sv_setiv',  '',      '(IV)',     '',    'POPi',    'SViv', ''     ],
 		'UV'      => ['UV',     'SvUV',      'sv_setuv',  '',      '(UV)',     '',    'POPu',    'SVuv', ''     ],
+		'Color'   => ['Color',  'SvUV',      'sv_setuv',  '',      '(Color)',  '',    'POPu',    'SVuv', ''     ],
 		'double'  => ['double', 'SvNV',      'sv_setnv',  '',      '(double)', '',    'POPn',    'SVnv', ''     ],
 		'char*'   => ['char *', 'SvPV_nolen','sv_setpv',  '(SV*)', '',         ', 0', 'POPp',    'SVpv', '' ],
 		'string'  => ['char',   'SvPV_nolen','sv_setpv',  '(SV*)', '',         ', 0', 'POPp',    'SVpv', '' ],
@@ -292,21 +266,21 @@ sub expect
 # expect given string
 {
 	my $tofind = shift;
-	error "APC001: Found ``$tok'' but expecting ``$tofind''" unless ( $tok = gettok) eq $tofind;
+	error "Found ``$tok'' but expecting ``$tofind''" unless ( $tok = gettok) eq $tofind;
 }
 
 sub check
 # check whether $tok is equal to given parm
 {
 	my $tofind = shift;
-	error "APC002: Found ``$tok'' but expecting ``$tofind''" unless $tok eq $tofind;
+	error "Found ``$tok'' but expecting ``$tofind''" unless $tok eq $tofind;
 }
 
 sub getid
 # waits for indentifier
 {
 	$tok = gettok;
-	error "APC003: Expecting identifier but found ``$tok''" unless $tok =~ /^\w+$/;
+	error "Expecting identifier but found ``$tok''" unless $tok =~ /^\w+$/;
 	return $tok;
 }
 
@@ -317,7 +291,7 @@ sub getidsym
 	$tok = gettok;
 	return $tok if index($_[0],$tok) >= 0;   # $what =~ /\Q$tok\E/;
 	return $tok if $tok =~ /^\w+$/;
-	error "APC004: Expecting identifier or one of ``$_[0]'' but found ``$tok''";
+	error "Expecting identifier or one of ``$_[0]'' but found ``$tok''";
 }
 
 sub getObjName
@@ -325,7 +299,7 @@ sub getObjName
 	$tok = gettok;
 	for ( split( '::', $tok))
 	{
-		error "APC033: Expecting object identifier but found ``$_''" 
+		error "Expecting object identifier but found ``$_''" 
 			unless $_ =~ /^\w+$/;
 	}
 	return $tok;
@@ -357,7 +331,7 @@ sub load_file
 		return;
 	}
 
-	open FILE, $fileName or die "APC009: Cannot open $fileName\n";
+	open FILE, $fileName or die "Cannot open $fileName\n";
 
 	# loading the file
 	my @_words;
@@ -382,7 +356,7 @@ sub load_file
 				push @_words, $1; push @_lineNums, $ln;
 			} else {
 				close FILE;
-				error "APC009a:  Invalid character in input";
+				error " Invalid character in input";
 			}
 			redo;
 		}
@@ -422,7 +396,7 @@ sub find_file
 		my %arraySave = %arrays;
 		my %defineSave = %defines;
 		my $nextStructureSave = $nextStructure;
-		eval { load_file( "$_$fileName") };
+		eval { load_file( "$_/$fileName") };
 		if ( !$@ && skipto( 'object'))
 		{
 			$tok = gettok;
@@ -439,7 +413,7 @@ sub find_file
 		%defines = %defineSave;
 		restore_context;
 		next unless $fit;
-		return "$_$fileName", $dyna;
+		return "$_/$fileName", $dyna;
 	}
 	return undef;
 }
@@ -447,7 +421,7 @@ sub find_file
 sub checkid
 {
 	my $id = $_[0];
-	error "APC017: Indentifier '$id' already defined or cannot be used; error"
+	error "Indentifier '$id' already defined or cannot be used; error"
 	if $mapTypes{$id} || $structs{$id} || $arrays{$id} || $defines{$id} || $forbiddenParms{$id};
 }
 
@@ -484,8 +458,8 @@ sub parse_struc
 	{
 		# determining typecasting
 		my $nid = getid;
-		error "APC051: Invalid casting type $id" unless ( $arrays{$nid} || $structs{$nid} || $defines{$nid});
-		error "APC052: Invalid forecasting type $id" if $redef;
+		error "Invalid casting type $id" unless ( $arrays{$nid} || $structs{$nid} || $defines{$nid});
+		error "Invalid forecasting type $id" if $redef;
 		expect(';');
 		if ( $arrays{$nid}) {
 			$arrays{ $id} = [ $arrays{$nid}[0], $arrays{$nid}[1], {%added}];
@@ -503,7 +477,7 @@ sub parse_struc
 			}
 			expect('[');
 			my $dim = gettok;
-			error "APC050: Invalid array $id dimension '$dim'" if $dim <= 0;
+			error "Invalid array $id dimension '$dim'" if $dim <= 0;
 			expect(']');
 			expect(';');
 			$arrays{ $id} = [ $dim, $type, {%added}];
@@ -520,24 +494,24 @@ sub parse_struc
 		last if $type eq "}";
 		if ( exists $structs{ $type}) {
 			$structure = $structs{ $type};
-			error "APC053: Cannot do nested arrays (yet)"
+			error "Cannot do nested arrays (yet)"
 				unless $structure-> [PROPS]-> {hash};
 			$type = 'SV*';
 		} else {
-			error "APC046: Invalid field type $type"
+			error "Invalid field type $type"
 				if (!defined $mapTypes{$type} || $type eq 'HV') && ( $type ne 'string');
 			$structure = $type = $mapTypes{$type} if $mapTypes{$type};
 		}
 		$tok = gettok;
 		# checking pointer
 		if ( $tok eq '*') {
-			error "APC045: Invalid pointer to $type in structure $id"
+			error "Invalid pointer to $type in structure $id"
 				if ( $type ne 'char') && ( $type ne 'SV');
 			$type .='*';
 			$structure = $type;
 			$tok = getid;
 		} else {
-			error "APC054: Invalid filed type $type in structure $id"
+			error "Invalid filed type $type in structure $id"
 				if ( $type eq 'HV') or ( $type eq 'SV');
 		}
 		push ( @types, $structure);
@@ -567,7 +541,7 @@ sub parse_struc
 			expect(';');
 		}
 	}
-	error "APC047: Structure $id defined as empty" unless scalar @ids;
+	error "Structure $id defined as empty" unless scalar @ids;
 
 	if ( $redef) {
 		# if it's hash default values redefine check for structure exact match
@@ -575,9 +549,9 @@ sub parse_struc
 		$ln = $lineStart;
 		my @cmpType = (@{$structs{$id}[0]});
 		my @cmpIds  = (@{$structs{$id}[1]});
-		error "APC048: Structure $id is already defined. Error" unless @cmpType eq @types;
+		error "Structure $id is already defined. Error" unless @cmpType eq @types;
 		for ( my $i = 0; $i < scalar @types; $i++) {
-			error "APC048: Structure $id is already defined. Error"
+			error "Structure $id is already defined. Error"
 			unless ( $cmpType[$i] eq $types[$i]) && ( $cmpIds[$i] eq $ids[$i]);
 		}
 		$ln = $lineStart;
@@ -602,7 +576,7 @@ sub parse_typedef
 		checkid( $t);
 		$type = $t;
 	}
-	error "APC049: Invalid casting type $type" unless ( $mapTypes{$type});
+	error "Invalid casting type $type" unless ( $mapTypes{$type});
 	expect(';');
 	my $incl = $fileName;
 	$incl =~ s/([^.]*)\..*$/$1/;
@@ -663,12 +637,12 @@ sub preload_method
 			$acc .= " " . $tok;
 			$checkEmptyRes .= $tok;
 		}
-		error "APC020: Function result haven't been defined" unless $checkEmptyRes;
-		error "APC012: Unexpected (" unless $id;
+		error "Function result haven't been defined" unless $checkEmptyRes;
+		error "Unexpected (" unless $id;
 	} else {
-		error "APC056: Property cannot be 'void'" if $proptype eq 'void';
+		error "Property cannot be 'void'" if $proptype eq 'void';
 	}
-	error "APC013: Method $id already defined"
+	error "Method $id already defined"
 		if exists($allMethodsHosts{ $id}) && $allMethodsHosts{ $id} eq $className;
 
 	@defParms = ();
@@ -731,16 +705,16 @@ sub preload_method
 
 			pop @gp;
 			$tok = shift @gp;
-			error "APC040: Invalid parameter $parmNo description" unless defined $tok;
+			error "Invalid parameter $parmNo description" unless defined $tok;
 			push ( @types, $tok); # first type description, definitive for
 										# default parameter validity check
 			if ( $tok eq '...') {
 				$ellipsis  = 1;
 				$hasEllipsis++;
-				error "APC057: Property can not contain ellipsis into parameter list" 
+				error "Property can not contain ellipsis into parameter list" 
 					if $asProperty;
 			} else {
-				error "APC043: Invalid parameter $parmNo description" 
+				error "Invalid parameter $parmNo description" 
 					unless defined $gp[0];
 			}
 			my $hasEqState = 0;
@@ -749,10 +723,10 @@ sub preload_method
 			{
 				# checking for type-name matching
 				checkid($_);
-				error "APC005: Indirect function and array parameters are unsupported. Error"
+				error "Indirect function and array parameters are unsupported. Error"
 					if /^[\(\)\[\]\&]$/;
-				error "APC006: Unexpected semicolon" if /^[\;]$/;
-				error "APC042: End of default parameter description expected but found $_" 
+				error "Unexpected semicolon" if /^[\;]$/;
+				error "End of default parameter description expected but found $_" 
 					if $hasEqState > 1;
 				if ( $hasEqState == 1)
 				{
@@ -761,20 +735,20 @@ sub preload_method
 				}
 				# default parameter add
 				if ( $_ eq "=") {
-					error "APC041: Invalid default parameter $parmNo description" 
+					error "Invalid default parameter $parmNo description" 
 						if $hasEqState;
-					error "APC058: Property can not contain default parameters" 
+					error "Property can not contain default parameters" 
 						if $asProperty;
 					$hasEqState = 1;
 				}
 			}
-			error "APC038: Default argument missing for parameter $parmNo"
+			error "Default argument missing for parameter $parmNo"
 				if !defined $defaultParm && defined $defParms[-1];
 			push( @defParms, $defaultParm);
 			last if ( $parm[-1] eq ")");
 			$parmNo++;
 		}
-		error "APC055: Ellipsis (...) could be defined only in the end of parameters list"
+		error "Ellipsis (...) could be defined only in the end of parameters list"
 			if $hasEllipsis > 1 || ( $hasEllipsis == 1 && $ellipsis == 0);
 	} else {
 		$acc .= "$proptype value" if $asProperty;
@@ -786,18 +760,18 @@ sub preload_method
 	$tok = gettok;
 	if ( $tok eq "=") {
 		$tok = gettok;
-		error "APC044: Expected => but found $tok" unless $tok eq '>';
+		error "Expected => but found $tok" unless $tok eq '>';
 		$piped = getid;
 		expect(';');
 	} else {
-		error "APC022: Expected semicolon or => but found $tok" unless $tok eq ';';
+		error "Expected semicolon or => but found $tok" unless $tok eq ';';
 	}
 
 	$acc .= ';';
 
 	# checking default parameters
 	for ( my $i = 0; $i <= $#types; $i++) {
-		error "APC0039: Parameter ${\($i+1)} of type $types[$i] cannot be used with default parameters"
+		error "Parameter ${\($i+1)} of type $types[$i] cannot be used with default parameters"
 		if defined $defParms[$i] && (
 			$structs{$types[$i]} || $arrays{$types[$i]} ||
 #        ( $types[$i] eq "Handle") || <<<< to enable being Handle default parameters
@@ -825,7 +799,7 @@ sub cmp_func
 sub store_method
 {
 	# comparing virtual overloaded methods bodies - die if unequal
-	error "APC016: Header '$id' does not match previous definition"
+	error "Header '$id' does not match previous definition"
 		if !exists($staticMethods{ $id}) &&
 			exists($allMethodsHosts{ $id}) && !($allMethodsHosts{ $id} eq $className) &&
 			($allMethodsHosts{ $id}) && !( cmp_func($allMethodsBodies[$allMethods{$id}], $acc));
@@ -871,7 +845,7 @@ sub parse_parms
 		my $valid = 0;
 		foreach (@lxmz)
 		{
-			error "APC019: Reserved word '$_' used" if $forbiddenParms{$_};
+			error "Reserved word '$_' used" if $forbiddenParms{$_};
 			$hvUsed = 1 if $_ eq "HV";
 		}
 		for ( my $j = 0; $j <= $#lxmz; $j++)
@@ -923,7 +897,7 @@ sub parse_parms
 
 	if ( $hvUsed) {
 		my @lxmz = split(" ", $chunks[ $#chunks]);
-		error "APC021: HV* can be declared only at the end of parameters list." unless $lxmz[ 0] eq "HV";
+		error "HV* can be declared only at the end of parameters list." unless $lxmz[ 0] eq "HV";
 	}
 	return $acc;
 }
@@ -948,7 +922,7 @@ sub load_single_part
 		$ownCType =~ s/^Prima:://;
 		$ownCType =~ s/::/__/g;
 	}
-	error "APC010: Class $className already defined" if $definedClasses{ $className};
+	error "Class $className already defined" if $definedClasses{ $className};
 	$definedClasses{ $className} = 1;
 	$tok = gettok;
 	$superClass = '';
@@ -967,7 +941,7 @@ sub load_single_part
 			}
 			expect ")";
 			$tok = gettok;
-			error "APC011: Cyclic inheritance detected.  Watch out ``$superClass''" 
+			error "Cyclic inheritance detected.  Watch out ``$superClass''" 
 				if $definedClasses{ $superClass};
 				
 			do {
@@ -980,15 +954,15 @@ sub load_single_part
 				putback( $tok);
 				my ( $superFile, $inhDyna) = find_file( "$superClass.cls");
 				$tok = gettok;
-				error "APC0035: Cannot find file corresponding to ''$superClass''" 
+				error "Cannot find file corresponding to ''$superClass''" 
 					unless defined $superFile;
-				error "APC0036: Static object cannot inherit dynamic module $superClass" 
+				error "Static object cannot inherit dynamic module $superClass" 
 					if !$genDyna && $inhDyna;
 				parse_file( $superFile) unless $sayparent;
 				$level--;
 			};
 		} else {
-			error "APC032: Cannot create dynamic object with no inheritance"
+			error "Cannot create dynamic object with no inheritance"
 			if $genDyna && ( $level == 0);
 		}
 	}
@@ -1026,7 +1000,7 @@ sub load_single_part
 				preload_method( 1, 0);
 				store_method;
 				unless ($level) {
-					error "APC018: Invalid types in import function $id declaration" 
+					error "Invalid types in import function $id declaration" 
 						unless parse_parms;
 					push (@portableImports, "${id}${acc}");
 					push (@newMethods, "${id}${acc}") unless $inherit;
@@ -1081,7 +1055,7 @@ sub load_single_part
 				$properties{$id} = $proptype;
 			} else {
 			# variable workaround
-				error "APC008: Expecting identifier but found ``$tok''" 
+				error "Expecting identifier but found ``$tok''" 
 					unless $tok =~ /^\w+$/;
 				$acc = $tok;
 				until (($tok = getidsym("*;[]()+")) eq ";")
@@ -1090,7 +1064,7 @@ sub load_single_part
 					$acc .= " " . $tok;
 				}
 				$acc .= ";";
-				error "APC015: Variable $id already defined" if $allVars{ $id};
+				error "Variable $id already defined" if $allVars{ $id};
 				$allVars{$id} = $className;
 				push(@allVars, $acc);
 			}
@@ -1143,7 +1117,14 @@ sub load_structures
 			my $fid = getid;
 			expect(';');
 			save_context;
-			load_file( "$fid.cls");
+			my $loaded;
+			foreach (@includeDirs) {
+				my $f = "$_/$fid.cls";
+				next unless -f $f;
+				$loaded = $f;
+			}
+			die "$fid.cls not found\n" unless defined $loaded;
+			load_file( $loaded);
 			$level++;
 			&load_structures;
 			$level--;
@@ -1162,13 +1143,15 @@ $tok = gettok;
 if (( $tok eq "object") || ( $tok eq "package")) {
 	load_single_part( $tok);
 } elsif ( $tok) {
-	error "APC037: found $tok but expecting 'object' or 'package'";
+	error "found $tok but expecting 'object' or 'package'";
 } else {
 	unless ( $level) {
 		#$genInc = 0;
 		$ownFile = $fileName;
 		$ownFile =~ s/([^.]*)\..*$/$1/;
-		$ownCType = uc $ownFile;
+		my $out = $ownFile;
+		$out =~ s[.*\/][];
+		$ownCType = uc $out;
 	}
 }
 } # end of parse_file
@@ -1932,12 +1915,14 @@ METHODS: 	for ( my $i = 0; $i < scalar @{$methods}; $i++) {
 
 
 	if ( $genH) {
-																	# 1 - class.h
-	open HEADER, ">$dirOut$ownFile.h" or die "APC009: Cannot open $dirOut$ownFile.h\n";
+
+	my $out = $ownFile;
+	$out =~ s[.*\/][];
+	open HEADER, ">$dirOut$out.h" or die "Cannot open $dirOut$out.h\n";
 print HEADER <<LABEL;
 /* This file was automatically generated.
- * Do not edit, you\'ll loose your changes anyway.
- * file: $ownFile.h  */
+ * Do not edit, you\'ll lose your changes anyway.
+ * file: $out.h  */
 #ifndef ${ownCType}_H_
 #define ${ownCType}_H_
 #ifndef _APRICOT_H_
@@ -1971,6 +1956,7 @@ print HEADER "#include \"$baseFile.h\"\n" if $baseClass;
 	}
 
 	for ( keys %toInclude) {
+		s[.*\/][];
 		print HEADER "#include \"$_.h\"\n";
 	}
 }
@@ -2138,13 +2124,15 @@ close HEADER;
 
 if ( $genInc) {
 																			#2 - class.inc
-open HEADER, ">$dirOut$ownFile.inc" or die "APC009: Cannot open $dirOut$ownFile.inc\n";
+my $out = $ownFile;
+$out =~ s[.*\/][];
+open HEADER, ">$dirOut$out.inc" or die "Cannot open $dirOut$out.inc\n";
 print HEADER <<LABEL;
 /* This file was automatically generated.
- * Do not edit, you\'ll loose your changes anyway.
- * file: $ownFile.inc */
+ * Do not edit, you\'ll lose your changes anyway.
+ * file: $out.inc */
 
-#include "$ownFile.h"
+#include "$out.h"
 LABEL
 print HEADER "#include \"guts.h\"\n" if ( !$genDyna && $genObject);
 print HEADER <<SD;
@@ -2323,11 +2311,13 @@ SD
 	} # end gen Inc
 
 if ( $genTml) {                                        #3 - class.tml
-	open HEADER, ">$dirOut$ownFile.tml" or die "APC009: Cannot open $dirOut$ownFile.tml\n";
+	my $out = $ownFile;
+	$out =~ s[.*\/][];
+	open HEADER, ">$dirOut$out.tml" or die "Cannot open $dirOut$out.tml\n";
 	print HEADER <<LABEL;
 /* This file was automatically generated.
- Do not edit, you'll loose your changes anyway.
- file: $ownFile.tml   */
+ Do not edit, you'll lose your changes anyway.
+ file: $out.tml   */
 
 LABEL
 
